
def generate_lammps_input(
    PKA_id=None,
    units=None,
    atom_style=None,
    read_data=None,
    masses=None,
    pair_styles=None,
    pair_coeffs=None,
    run_steps=None
):
    """
    Generates a LAMMPS input file (tde.in) from a set of parameters.

    Parameters:
        PKA_id (int): ID of the PKA atom
        units (str): Units for the simulation (default: 'metal')
        atom_style (str): Atom style for the simulation (default: 'atomic')
        read_data (str): Filename of the data file to read in (default: 'structure.data')
        masses (list of tuples): List of tuples containing the atom type and mass for each atom
        pair_styles (list of str): List of pair styles to use in the simulation
        pair_coeffs (list of str): List of pair coefficients to use in the simulation
        run_steps (int): Number of steps to run in the simulation (default: 2000)

    Returns:
        None
    """
    if masses is None:
        print('ERROR: Must specify masses for atoms (list of tuples [(atom_type, mass),...])')
        exit()
    if pair_styles is None:
        print('ERROR: Must specify pair styles (list of strings)')
        exit()
    if pair_coeffs is None:
        print('ERROR: Must specify pair coeffs (list of strings)')
        exit()
    if PKA_id is None:
        print('ERROR: Must specify PKA id (int)')
        exit()
    if read_data is None:
        print('ERROR: Must specify data file (str)')
        exit()
    if run_steps is None:   
        print('ERROR: Must specify run steps (int)')
        exit()
    
    if units is None:
        print(f"ERROR: Must specify units (str)")
        exit()

    if atom_style is None:
        print(f"ERROR: Must specify atom_style (str)")
        exit()

    lines = []
    # Basic setup
    lines.append(f"units {units}")
    lines.append(f"dimension 3")
    lines.append(f"boundary p p p")
    lines.append("")
    lines.append(f"atom_style {atom_style}")
    lines.append("")
    lines.append(f"read_data ../{read_data}")
    lines.append("")

    # Masses
    for atom_type, mass in masses:
        lines.append(f"mass {atom_type} {mass}")
    lines.append("")

    # Pair styles
    for ps in pair_styles:
        lines.append(f"{ps}")
    lines.append("")
    for pc in pair_coeffs:
        lines.append(f"{pc}")
    lines.append("")

    lines.extend([
        "neigh_modify delay 0 every 1 check yes",
        "",
        f"thermo_style custom step temp pe ke press vol cella cellb cellc",
        f"thermo 10",
        ""
    ])

    # Box and outer shell
    lines.extend([
        "# Define the entire simulation box dimensions",
        "variable xlo equal bound(all,xmin)",
        "variable xhi equal bound(all,xmax)",
        "variable ylo equal bound(all,ymin)",
        "variable yhi equal bound(all,ymax)",
        "variable zlo equal bound(all,zmin)",
        "variable zhi equal bound(all,zmax)",
        "",
        "# Define variables for the outer shell boundaries",
        "variable xlo_outer equal ${xlo} + 2.5",
        "variable xhi_outer equal ${xhi} - 2.5",
        "variable ylo_outer equal ${ylo} + 2.5",
        "variable yhi_outer equal ${yhi} - 2.5",
        "variable zlo_outer equal ${zlo} + 2.5",
        "variable zhi_outer equal ${zhi} - 2.5",
        "",
        "# Define regions for the outer shell",
        "region outer_xlo block ${xlo} ${xlo_outer} INF INF INF INF",
        "region outer_xhi block ${xhi_outer} ${xhi} INF INF INF INF",
        "region outer_ylo block INF INF ${ylo} ${ylo_outer} INF INF",
        "region outer_yhi block INF INF ${yhi_outer} ${yhi} INF INF",
        "region outer_zlo block INF INF INF INF ${zlo} ${zlo_outer}",
        "region outer_zhi block INF INF INF INF ${zhi_outer} ${zhi}",
        "",
        "# Combine the regions for the outer shell",
        "region outer_shell union 6 outer_xlo outer_xhi outer_ylo outer_yhi outer_zlo outer_zhi",
        "",
        "# Group atoms in the outer shell",
        "group boundary region outer_shell",
        "",
        "group core subtract all boundary",
        f"group PKA id {PKA_id}",
        "",
        f"velocity PKA set MYPKAVEL",
        "fix core_fix all nve",
        "fix boundary_fix boundary langevin 0 0 0.1 48279",
        f"dump positions all atom {run_steps} prod.data",
        "dump_modify positions first yes",
        "",
        f"run {run_steps}"
    ])

    # Write to file
    with open('tde.in', "w") as f:
        f.write("\n".join(lines))

    print(f"LAMMPS input file 'tde.in' generated successfully!")




import numpy as np
from scipy.interpolate import griddata
import matplotlib
matplotlib.use("Agg")   # headless, no GUI
import matplotlib.pyplot as plt
from typing import Tuple, Optional, Callable, List, Literal
from sphere_utils import sphere_to_polar

class Plot_Tools:
    

    def __init__(self, input_points, output_values, theta_range, phi_range):
        
        self.input_points = input_points
        self.output_values = output_values
        self.theta_range = theta_range
        self.phi_range = phi_range
   

    
    def plot(self, plot_type: Literal["var", "mean"], fname=None):

        """
        Plot a heatmap of the TDE surface in polar coordinates.

        Parameters:
            plot_type (Literal["var", "mean"]): Type of plot to generate. If "mean", plot the mean TDE. If "var", plot the variance of the TDE.
            fname (Optional[str]): If not None, name of the output file to save the plot to.

        Returns:
            None
        """
        theta, phi = sphere_to_polar(self.input_points)

         # Enforce periodicity so plot works correctly
        phi_extended = np.concatenate([phi, phi - 2*np.pi, phi + 2*np.pi])
        # Repeat theta and output to match size of periodic phi
        theta_extended = np.tile(theta, 3)
        values_extended = np.tile(self.output_values, 3)

        # phi_grid, theta_grid, grid_values, phi, theta = self.generate_sphere_data(input=input_points, output=output_values)
        n_phi, n_theta = 300, 150
        phi_grid = np.linspace(self.phi_range[0], self.phi_range[1], n_phi)
        theta_grid = np.linspace(self.theta_range[0], self.theta_range[1], n_theta)
        phi_mesh, theta_mesh = np.meshgrid(phi_grid, theta_grid)

        grid_values = griddata(
            (phi_extended, theta_extended),
            values_extended,
            (phi_mesh, theta_mesh),
            method='linear'
        )

        # Plot on polar projection
        fig, ax = plt.subplots(subplot_kw=dict(projection='polar'))
       

        # Apply limits (convert θ to degrees since you plot np.degrees(theta))
        ax.set_rlim(np.degrees(self.theta_range[0]), np.degrees(self.theta_range[1]))
        ax.set_thetalim(self.phi_range[0], self.phi_range[1])
        
        
        ax.set_theta_zero_location("E")
        
        c = ax.pcolormesh(phi_grid, np.degrees(theta_grid), grid_values,
                        cmap='plasma' if plot_type == "mean" else 'viridis', shading='gouraud')
        # Add contour lines on top
        contour_levels = np.linspace(0, np.max(self.output_values), 10)  # 8 contour lines
        cs = ax.contour(
             phi_grid, np.degrees(theta_grid), grid_values,
             levels=contour_levels, colors='k', linewidths=0.5, alpha=0.7
        )
        #ax.clabel(cs, inline=True, fontsize=6, fmt="%.0f")  # label contour values

        cb = fig.colorbar(c, ax=ax, label="E$_d$ (eV)" if plot_type == "mean" else r"$\sigma^2$ (eV$^2$)", pad=0.1)
    

        # ax.scatter(phi, np.degrees(theta), c='black', s=0.1)
       
        plt.savefig(('tde_map.png' if plot_type == "mean" else 'tde_map_var.png') if fname is None else fname, dpi=600)
        # plt.show()





import torch
import gpytorch
import numpy as np
import os 
from math import pi

def optimise_model (model, X_train, y_train_std, likelihood, mll):
    # Optimizer (Adam with all model parameters)
    """
    Optimizes a Gaussian Process model using Adam optimizer.

    Args:
        model (ExactGPModel): The Gaussian Process model to be optimized.
        X_train (torch.Tensor): The training inputs.
        y_train_std (torch.Tensor): The standardized training outputs.
        likelihood (MultivariateNormal): The likelihood of the training outputs.
        mll (gpytorch.mlls.ExactMarginalLogLikelihood): The marginal log likelihood of the training outputs.

    Returns:
        None
    """
    optimizer = torch.optim.Adam(model.parameters(), lr=0.1)

    # Train mode for model + likelihood
    model.train(); likelihood.train()
    prev_loss = None
    for i in range(10000):
        optimizer.zero_grad()
        output = model(X_train)
        loss = -mll(output, y_train_std)
        loss_value = loss.item()
        loss.backward()
        optimizer.step()
        # Early stopping based on tolerance
        if prev_loss is not None and abs(prev_loss - loss_value) < 0.001:
            # print(f"Converged at iteration {i} with change {abs(prev_loss - loss_value):.6e}")
            break

        
        prev_loss = loss_value

def converge_check(imse, patience=30, slope_tol=1e-3):
    """
    More robust convergence check based on IMSE stabilization.

    Parameters:
        imse (list or np.ndarray): Sequence of IMSE values.
        patience (int): Number of recent iterations to consider.
        slope_tol (float): Maximum allowed average slope magnitude of recent IMSEs.

    Returns:
        bool: True if converged, False otherwise.
    """
    if len(imse) < patience:
        return False

    recent = np.array(imse[-patience:])

   

    # Compute average slope (simple linear regression)
    x = np.arange(len(recent))
    coeffs = np.polyfit(x, recent, 1)
    slope = coeffs[0]

    # Converged if slope near zero
    if abs(slope) < slope_tol:
        return True
    return False


def query_model(model, probe_pts, y_std, y_mean,likelihood=None):
    
    """
    Queries a Gaussian Process model at a set of probe points.

    Args:
        model (ExactGPModel): The Gaussian Process model to be queried.
        probe_pts (torch.Tensor): The points at which to query the model.
        y_std (float): The standard deviation of the training outputs.
        y_mean (float): The mean of the training outputs.
        likelihood (MultivariateNormal): The likelihood of the training outputs.

    Returns:
        mean_pred (torch.Tensor): The predicted mean values at the probe points.
        var_pred (torch.Tensor): The predicted variance values at the probe points.
    """
    if likelihood is not None:
        likelihood.eval()
    model.eval()
    with torch.no_grad():
        if likelihood is not None:
            preds = likelihood(model(probe_pts), noise=torch.full((probe_pts.shape[0],), 0.5))
        else:
            preds = model(probe_pts)
        mean_pred = preds.mean * y_std + y_mean     # de-standardize back to eV
        var_pred = preds.variance * (y_std**2)

    return mean_pred, var_pred


def save_checkpoint(X_train, y_train, noise_train, it, imse, stage="iter"):
    """
    Saves a checkpoint of the current model state.

    Args:
        X_train (torch.Tensor): The training inputs.
        y_train (torch.Tensor): The training outputs.
        noise_train (torch.Tensor): The training noise.
        it (int): The current iteration number.
        imse (float): The current IMSE value.
        stage (str): The current stage of the active learning loop.

    Returns:
        None
    """
    torch.save({
        'X_train': X_train,
        'y_train': y_train,
        'noise_train': noise_train,
        'iteration': it,
        'stage': stage,
        'imse': imse
    }, 'checkpoint.pt')

def load_checkpoint():
    """
    Loads a checkpoint of the current model state.

    Returns:
        dict: A dictionary containing the state of the model if a checkpoint exists, None otherwise.
    """
    if os.path.exists('checkpoint.pt'):
        return torch.load('checkpoint.pt', weights_only=False)
    return None

def estimate_imse(model, likelihood, probe_pts, y_std, theta_range, phi_range):
    """
    Estimates the integrated mean squared error (IMSE) of the model on a given set of probe points.

    Parameters:
        model (ExactGPModel): The Gaussian Process model to estimate the IMSE for.
        likelihood (MultivariateNormal): The likelihood of the training outputs.
        probe_pts (torch.Tensor): The probe points to estimate the IMSE at.
        y_std (torch.Tensor): The standard deviation of the training outputs.
        theta_range (tuple[float, float], optional): The range of theta values to estimate the IMSE for. Defaults to None.
        phi_range (tuple[float, float], optional): The range of phi values to estimate the IMSE for. Defaults to None.

    Returns:
        float: The estimated IMSE value.
    """
    model.eval(); likelihood.eval()
    if theta_range is not None:
        theta_min, theta_max = theta_range
    else:
        theta_min, theta_max = 0, np.pi
    if phi_range is not None:
        phi_min, phi_max = phi_range
    else:
        phi_min, phi_max = 0, 2 * np.pi
    # Fraction of total surface area of sphere (to scale IMSE expression)
    area_fraction = (phi_max - phi_min) * (np.cos(theta_min) - np.cos(theta_max)) / (4 * np.pi)
    
    # Don't compute gradients for resources, use fater vairance prediction at minimal accuracy loss
    with torch.no_grad():
        # Extract model predictions at probe points
        preds = model(probe_pts)
        # Preds is GPyTorch MultivariateNormal object with .mean and .variance
        # Multiply by train_y_std^2 to get variance in original units
        var = torch.clamp(preds.variance, min=0.0) * (y_std**2)
    # Integrating over sphere requires 4 pi multiplier so variance spread over the whole sphere gives higher value.
    return area_fraction * ((4*pi )/ probe_pts.shape[0]) * var.sum().item() # Sum of vairance at all points, multiplied by weight factor (float)




__version__ = "0.3.0"
from DEmap.config import Config
from DEmap.main import demap
from DEmap.lmp_input_generator import generate_lammps_input

__all__ = ["Config", "demap", "generate_lammps_input"]





import numpy as np
from math import pi
from typing import Tuple, Optional, Callable, List




def sph_to_unit(theta, phi):
    """Convert spherical coordinates (theta colatitude, phi azimuth) to unit vectors."""
    x = np.sin(theta) * np.cos(phi)
    y = np.sin(theta) * np.sin(phi)
    z = np.cos(theta)
    return np.stack([x, y, z], axis=1)


def fibonacci_sphere(
    n: int,
    theta_range: Optional[Tuple[float, float]] = None,
    phi_range: Optional[Tuple[float, float]] = None,
) -> np.ndarray:
 
    """
    Generate a set of points evenly distributed on the surface of a unit sphere.

    Parameters:
        n (int): The number of points to generate.
        theta_range (Optional[Tuple[float, float]], optional): The range of theta values to generate points in.
        phi_range (Optional[Tuple[float, float]], optional): The range of phi values to generate points in.

    Returns:
        np.ndarray: The set of points evenly distributed on the surface of a unit sphere.
    """
    if n < 50:
        UserWarning('Regression is unstable for low initial point counts -> Use init_pts greater than 50!')
        exit()
    
    # Determine the percentage of the sphere covered by given ranges
    if theta_range is not None:
        theta_min, theta_max = theta_range
    else:
        theta_min, theta_max = 0, np.pi
    if phi_range is not None:
        phi_min, phi_max = phi_range
    else:
        phi_min, phi_max = 0, 2 * np.pi
    area_fraction = (phi_max - phi_min) * (np.cos(theta_min) - np.cos(theta_max)) / (4 * np.pi)
    
    n = n * (1/area_fraction)

    # Epsilon is an offset to help counter clustering towards poles 
    epsilon = 0.5  
    i = np.arange(n)
    phi = (1 + 5 ** 0.5) / 2
    # golden angle
    ga = 2 - 2 / phi
    
    z = 1 - 2 * (i + epsilon) / n
    z = np.clip(z, -1.0, 1.0)
    r = np.sqrt(1 - z * z)
    theta =  2 * np.pi * ((i * ga) % 1)

    pts = np.stack([r * np.cos(theta), r * np.sin(theta), z], axis=1)
    pts /= np.linalg.norm(pts, axis=1, keepdims=True)
    
    theta, phi = sphere_to_polar(pts)

     # Apply masks
    mask = np.ones(len(pts), dtype=bool)
    if theta_range is not None:
        mask &= (theta >= theta_range[0]) & (theta <= theta_range[1])
    if phi_range is not None:
        mask &= (phi >= phi_range[0]) & (phi <= phi_range[1])
    

    return pts[mask]

def random_sphere_points(
    n: int,
    theta_range: tuple[float, float] = (0, np.pi),
    phi_range: tuple[float, float] = (0, 2 * np.pi)
) -> np.ndarray:
    
    """
    Generates n random points distributed on the surface of a sphere.

    Parameters:
    n (int): Number of points to generate.
    theta_range (tuple[float, float], optional): Range of theta (colatitude) values to generate. Defaults to (0, π).
    phi_range (tuple[float, float], optional): Range of phi (azimuth) values to generate. Defaults to (0, 2π).

    Returns:
    np.ndarray: Array of shape (n, 3) containing the generated points in Cartesian coordinates.
    """
    theta_min, theta_max = theta_range
    phi_min, phi_max = phi_range

    # Sample uniformly on the sphere’s surface
    u = np.random.rand(n)
    v = np.random.rand(n)
    
    # Convert to θ, φ ensuring uniform area distribution
    theta = np.arccos(np.cos(theta_min) - u * (np.cos(theta_min) - np.cos(theta_max)))
    phi = phi_min + (phi_max - phi_min) * v

    # Convert spherical to Cartesian coordinates
    x = np.sin(theta) * np.cos(phi)
    y = np.sin(theta) * np.sin(phi)
    z = np.cos(theta)

    return np.stack((x, y, z), axis=1)

def sphere_to_polar(points):
    # Points are (N, 3) on unit sphere
    """
    Convert points on the surface of a unit sphere to polar coordinates.

    Parameters:
        points (np.ndarray): Array of shape (N, 3) containing the points on the surface of a unit sphere.

    Returns:
        tuple: A tuple containing the polar angle (theta) and azimuthal angle (phi) of the points.
    """
    x, y, z = points[:, 0], points[:, 1], points[:, 2]
    r = np.linalg.norm(points, axis=1)
    theta = np.arccos(z / r)       # polar angle [0, pi]
    phi = np.arctan2(y, x)         # azimuthal angle [-pi, pi]
    phi = np.mod(phi, 2 * np.pi)   # convert to [0, 2pi)
    return theta, phi



import torch
import gpytorch

#--- Define custom Geodesic Kernel ---
class GeodesicMaternKernel(gpytorch.kernels.Kernel):
    is_stationary = True

    def __init__(self, lengthscale=1.0, nu=0.5, **kwargs):
        super().__init__(**kwargs)
        self.nu = nu
        self.register_parameter(name="raw_lengthscale", parameter=torch.nn.Parameter(torch.tensor(lengthscale)))
        self.register_constraint("raw_lengthscale", gpytorch.constraints.Positive())

    @property
    def lengthscale(self):
        return self.raw_lengthscale_constraint.transform(self.raw_lengthscale)

    @lengthscale.setter
    def lengthscale(self, value):
        self._set_lengthscale(value)

    def forward(self, x1, x2, diag=False, **params):
        # normalize inputs to lie on unit sphere
        x1_norm = x1 / x1.norm(dim=-1, keepdim=True)
        x2_norm = x2 / x2.norm(dim=-1, keepdim=True)

        # geodesic distance
        cos_theta = torch.matmul(x1_norm, x2_norm.transpose(-2, -1)).clamp(-1.0, 1.0)
        dist = torch.acos(cos_theta)

        # Matern kernel (nu=3/2)
        if self.nu == 1.5:
            sqrt3_d = torch.sqrt(torch.tensor(3.0)) * dist / self.lengthscale
            covar = (1.0 + sqrt3_d) * torch.exp(-sqrt3_d)
        elif self.nu == 2.5:
            sqrt5_d = torch.sqrt(torch.tensor(5.0)) * dist / self.lengthscale
            covar = (1.0 + sqrt5_d + 5.0*dist**2/(3.0*self.lengthscale**2)) * torch.exp(-sqrt5_d)
        elif self.nu == 0.5:
            covar = torch.exp(-dist / self.lengthscale)
        else:
            # general case using Bessel function (less common)
            raise NotImplementedError("Only nu=0.5, 1.5, 2.5 are implemented in closed form.")

        if diag:
            return covar.diag()
        else:
            return covar

# Exact GP object using Geodesic matern
class ExactGPModel(gpytorch.models.ExactGP):
    def __init__(self, train_x, train_y, likelihood):
        super().__init__(train_x, train_y, likelihood)
        self.mean_module = gpytorch.means.ConstantMean()
        self.covar_module = GeodesicMaternKernel(lengthscale=1.0, nu=1.5)

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)





from dataclasses import dataclass


# CFG params for running DEmap, contains deafults.
@dataclass
class Config:
    run_line: str
    init_n: int = 16
    probe_n: int = 10000
    max_iters: int = 10
    random_seed: int = 123



"""
TDE Surface Active Learning with GPyTorch on the Sphere
-----------------------------------------------------

- Models TDE as a function on the unit sphere using a Gaussian Process.
- Uses GPyTorch for scalable GP inference.
- Space-filling initialization with Fibonacci lattice.
- Active learning loop that reduces global surface uncertainty via IMSE.
"""


import numpy as np
from DEmap import __version__
from typing import Tuple, Optional
import torch
import gpytorch
from plot_tools import *
from TDE import evaluate_tde
from sphere_utils import fibonacci_sphere, random_sphere_points
from config import Config
from kernels import ExactGPModel
from gp_utils import save_checkpoint, load_checkpoint, query_model, estimate_imse, optimise_model, converge_check
from lmp_input_generator import generate_lammps_input




def demap(cfg, theta_range: Optional[Tuple[float, float]] = None,
    phi_range: Optional[Tuple[float, float]] = None, restart: bool = False,
    plot: bool=False):
    """
    Active learning of the TDE surface using Gaussian Processes.

    Parameters:
    cfg (Config): Config object containing hyperparameters for the run
    theta_range (Optional[Tuple[float, float]]): Range of theta values for the sphere
    phi_range (Optional[Tuple[float, float]]): Range of phi values for the sphere
    restart (bool): If True, load the latest checkpoint and continue from there
    plot (bool): If True, plot the mean and variance of the TDE on all probe points

    Returns:
    None
    """


    startup_text = f'''
---------------------------------------------
  ___  ___                
 |   \| __|_ __  __ _ _ __ 
 | |) | _|| '  \/ _` | '_ \ 
 |___/|___|_|_|_\__,_| .__/ 
                     |_|

VERSION: {__version__} 
AUTHOR: Ashley Dickson
CORRESPONDANCE: a.dickson2@lancaster.ac.uk
---------------------------------------------
    ''' 
    print(startup_text)
    # Validate theta_range
    if theta_range is not None:
        if not (0 <= theta_range[0] < theta_range[1] <= np.pi):
            raise ValueError(
                f"Invalid theta_range {theta_range}. "
                f"It must satisfy 0 <= theta_min < theta_max <= pi."
            )

    # Validate phi_range
    if phi_range is not None:
        if not (0 <= phi_range[0] < phi_range[1] <= 2 * np.pi):
            raise ValueError(
                f"Invalid phi_range {phi_range}. "
                f"It must satisfy 0 <= phi_min < phi_max <= 2pi."
            )
    # Initialise seeds to maintain stability on restart   
    np.random.seed(cfg.random_seed)
    torch.manual_seed(cfg.random_seed)
    # Load checkpoint file on restart 
    checkpoint = load_checkpoint() if restart else None
    # Initialise list of IMSE for each iteration
    imse_list = []
    if checkpoint and checkpoint['stage'] == "init":
        # Logic for restarting on initiasl point sample
        X_train = checkpoint['X_train']
        y_train = checkpoint['y_train']
        noise_train = checkpoint['noise_train']
        start_idx = X_train.shape[0]
        init_pts = fibonacci_sphere(n=cfg.init_n, theta_range=theta_range, phi_range=phi_range)
        print(f"Resuming initial sampling at point {start_idx}/{len(init_pts)}")
    elif not checkpoint:
        # Run from scratch w/o restart 
        print('Starting initial point sampling...')
        init_pts = torch.tensor(fibonacci_sphere(n=cfg.init_n, theta_range=theta_range, phi_range=phi_range), dtype=torch.float32)
        print('Total number of initial points:', len(init_pts))
        X_train = torch.empty((0, len(init_pts[0])), dtype=torch.float32)
        y_train = torch.empty((0,), dtype=torch.float32)
        noise_train = torch.empty((0,), dtype=torch.float32)
        start_idx = 0
    else:
        # Loaded full checkpoint (past init)
        print('Resuming from checkpoint.pt...')
        X_train = checkpoint['X_train']
        y_train = checkpoint['y_train']
        noise_train = checkpoint['noise_train']
        start_idx = None

    # Initial sampling 
    if start_idx is not None:
        for i in range(start_idx, len(init_pts)):
            u = init_pts[i]
            val, noise = evaluate_tde(cfg=cfg, u=u)
            print(f"TDE of initial point {i+1}/{len(init_pts)}: {val:.2f} eV")

            X_train = torch.cat([X_train, u.unsqueeze(0).clone().detach()])
            y_train = torch.cat([y_train, torch.tensor([val])])
            noise_train = torch.cat([noise_train, torch.tensor([noise])])

            save_checkpoint(X_train, y_train, noise_train, i, imse_list, stage="init")

        print('Initial points sampled...')

    # Build GP model 
    y_mean = y_train.mean()
    y_std = y_train.std()
    # Standardise to unit variance
    y_train_std = (y_train - y_mean) / y_std

    # Likelihood (noise) model
    likelihood = gpytorch.likelihoods.FixedNoiseGaussianLikelihood(
        noise=noise_train, learn_additional_noise=False
    )
    model = ExactGPModel(X_train, y_train_std, likelihood)
    mll = gpytorch.mlls.ExactMarginalLogLikelihood(likelihood, model)

    # Optimise model hyperparams given initial data 
    optimise_model(model, X_train, y_train_std, likelihood, mll)

    # Generate probe points on sphere (randomly distributed)
    probe = torch.tensor(random_sphere_points(n=cfg.probe_n, theta_range=theta_range, phi_range=phi_range), dtype=torch.float32)
    
    # Iterative loop 
    start_it = checkpoint['iteration'] + 1 if checkpoint and checkpoint['stage'] == "iter" else 1
    
    for it in range(start_it, cfg.max_iters + 1):
        
        imse = estimate_imse(model, likelihood, probe, y_std, theta_range, phi_range)
        imse_list.append(imse)
        if converge_check(imse_list):
            print('IMSE converged, exiting...')
            break

        save_checkpoint(X_train, y_train, noise_train, it, imse, stage="iter")
        
        # Grab predicted variance at all probe points
        _, var_pred = query_model(model=model, likelihood=None, probe_pts=probe, y_std=y_std, y_mean=y_mean)
        idx = torch.argmax(var_pred).item()
        # Find point with maximum variance
        new_pts = probe[idx].unsqueeze(0)

        pred_mean, pred_var = query_model(model=model, likelihood=None, probe_pts=new_pts, y_std=y_std, y_mean=y_mean)
     
        
        # Efficient TDE evaluation by inclusion of guess start energy two standard deviations below predicted TDE, if lower than 1 eV
        # otherwise start from 1 eV
        guess_start_energy = max(int(round(pred_mean[0].item() - (pred_var[0].sqrt().item() * 2))), 1)
        # New TDE and new Noise
        new_y, new_n = [], []
        
        val, noise = evaluate_tde(cfg=cfg,u=new_pts[0], start_energy=guess_start_energy)
        if it == start_it:
            print(f"{'Iter':>5} | {'Evaluated TDE (eV)':>20} | {'Predicted TDE (eV)':>22} | {'IMSE':>8}")
            print("-" * 70)

        # Print nice summary
        print(
            f"{it:5d} | "
            f"{val:20.2f} | "
            f"{pred_mean[0].item():10.2f} ± {np.sqrt(pred_var[0].item())*2:6.2f} eV | "
            f"{imse:8.2f}"
        )
        new_y.append(val)
        new_n.append(noise)
        # Add new data to train arrays
        X_train = torch.cat([X_train, new_pts])
        y_train = torch.cat([y_train, torch.tensor(new_y)])
        noise_train = torch.cat([noise_train, torch.tensor(new_n)])
        likelihood.noise = noise_train.clone()
        # Renormalise data 
        y_mean = y_train.mean()
        y_std = y_train.std()
        y_train_std = (y_train - y_mean) / y_std

        # Retrain model
        model.set_train_data(X_train, y_train_std, strict=False)
        optimise_model(model, X_train, y_train_std, likelihood, mll)
        # Create arrays for evaluations at probe points
        # mean_pred, var_pred = query_model(model=model, likelihood=likelihood, probe_pts=probe, y_std=y_std, y_mean=y_mean)
        
        # np.savetxt('probe_points.txt', probe.numpy())
        # np.savetxt('tde_points.txt', mean_pred.numpy())
        # np.savetxt('var_points.txt', var_pred.numpy())


    # Evaluate trained GP on all probe points
    mean_pred, var_pred = query_model(model=model, likelihood=likelihood, probe_pts=probe, y_std=y_std, y_mean=y_mean)

    X_train_np = probe.numpy()
    train_y_np = mean_pred.numpy()
    # Save vectors
    np.savetxt('probe_points.txt', probe.numpy())
    # Save TDE in eV
    np.savetxt('tde_points.txt', mean_pred.numpy())
    # Save variance in eV**2
    np.savetxt('var_points.txt', var_pred.numpy())
    # If plot argument is used, plot on all probe points
    if plot:
        #Plot TDE
        Plot_Tools(X_train_np, train_y_np, theta_range=theta_range, phi_range=phi_range).plot(plot_type="mean", fname='TDEMAP.png')
        #Plot variance
        Plot_Tools(X_train_np, var_pred.numpy(), theta_range=theta_range, phi_range=phi_range).plot(plot_type="var", fname='VARMAP.png')
    

 

# if __name__ == "__main__":
#     generate_lammps_input(masses = [(1, 55.845)], pair_styles = ['pair_style hybrid/overlay eam/fs tabgap'], 
#                            pair_coeffs = ['pair_coeff * * eam/fs ../Fe-2021-04-06.eam.fs Fe', 
#                                           'pair_coeff * * tabgap ../Fe-2021-04-06.tabgap Fe no yes'],
#                              units = 'metal', atom_style = 'atomic', read_data = 'Fe.data', run_steps = 2000, PKA_id=4367)

#     cfg = Config(run_line='mpirun /storage/hpc/51/dickson5/codes/tablammps/lammps_w_hdf5/build/lmp -in tde.in',
#                                init_n=200, max_iters=1400)

#     demap(theta_range=(0, np.pi/4), phi_range=(0, np.pi/4), restart=False, cfg=cfg, plot=False)




import numpy as np
from ovito.io import *
from ovito.modifiers import *
from ovito.pipeline import *
import numpy as np
import subprocess
import random
import os
from config import Config
from typing import Tuple, Optional, Callable, List

class TDE_simulation:

    def __init__(self, direction=None, run_line=None):

        # direction is np array 
        """
        Initialises a TDE_simulation object.

        Parameters:
            direction (np.ndarray): Direction of the TDE simulation
            run_line (str): LAMMPS command to run the simulation

        Attributes:
            direction (np.ndarray): Direction of the TDE simulation
            mass_data (dict): Masses of each atom type
            atom_id (int): ID of the PKA atom
            lammps_data_file (str): Filename of the LAMMPS data file
            pka_mass (float): Mass of the PKA atom
            run_line (str): LAMMPS command to run the simulation
        """
        self.direction = direction

        mass_data  = {}
        with open('tde.in', 'r') as f:
            lines = f.readlines()
        for line in lines:
            if 'mass' in line:
                _, atom_type, mass = line.split()
                mass_data[int(atom_type)] = float(mass)
            if 'PKA id' in line:
                atom_id = int(line.split()[-1])
            if 'read_data' in line:
                self.lammps_data_file = line.split()[-1]
 

        self.mass_data = mass_data
        

        # ID of PKA atom
        self.atom_id = atom_id
        if not isinstance(self.atom_id, int):
            raise TypeError('atom_id must be an integer')
        

        # Initialise PKA mass at start of calcualtions
        self.get_pka_mass()

        self.run_line = run_line




    def get_pka_mass (self):
        """
        Retrieves the mass of the PKA atom from the LAMMPS data file.
        """
        with open (self.lammps_data_file.strip('../'), 'r') as f:
            lines = f.readlines()
        for index, line in enumerate(lines):
            if 'Atoms' in line:
                atom_data_start = index + 2
        data_lines = lines[atom_data_start:]

        # Parse data for PKA atom
        for line in data_lines:
            atom_data = line.split()
            if int(atom_data[0]) == int(self.atom_id):
                atom_type = atom_data[1]
                break
    
        self.pka_mass =  self.mass_data[int(atom_type)]



    def calculate_pka_vel (self, pka_vector, pka_energy):
        """
        Calculates the velocity vector of the PKA atom given its direction and energy.

        Parameters:
            pka_vector (list of floats): direction of the PKA atom
            pka_energy (float): energy of the PKA atom in eV

        Returns:
            pka_velocity_vector (list of floats): velocity vector of the PKA atom in lammps units
        """
        eV_to_J = 1.60218e-19  # J/eV
        amu_to_kg = 1.66e-27   # kg/amu

        pka_vector_magnitude = np.linalg.norm(pka_vector)
        unit_vector = pka_vector / pka_vector_magnitude


        pka_velocity = np.sqrt(2 * pka_energy * eV_to_J / (self.pka_mass * amu_to_kg))
        pka_velocity = pka_velocity/100 #in lammps units

        pka_velocity_vector = pka_velocity * unit_vector

        # print(f"Velocity magnitude: {pka_velocity:.2f} m/s")
        # print(f"Velocity vector: {pka_velocity_vector}")

        # pka velocity vector is list of floats
        return pka_velocity_vector

    def defect_check (self):

        # print('Checking for defects..')
        """
        Checks the simulation for defects.

        Returns:
            int: The total number of defects (antisite defects + vacancies + interstitial defects)
        """
        filename = 'prod.data'
        pipeline = import_file(filename)


        ws_modifier = WignerSeitzAnalysisModifier(
            per_type_occupancies = True
            #eliminate_cell_deformation = True,
            #affine_mapping = ReferenceConfigurationModifier.AffineMapping.ToReference
        )


        pipeline.modifiers.append(ws_modifier)
        for frame in range(1, pipeline.source.num_frames):
            data = pipeline.compute(frame)
            occupancies = data.particles['Occupancy'].array
            occupancy2 = 0 #total num interstitial
            occupancy0 = 0 #total num vacancies
            # Get the site types as additional input:
            site_type = data.particles['Particle Type'].array
            # Calculate total occupancy of every site:
            try:
                total_occupancy = np.sum(occupancies, axis=1)
            except np.AxisError:
                total_occupancy = occupancies
            #print(total_occupancy)
            for element in total_occupancy:
                if element == 0:
                    occupancy0 +=1
                if element >= 2:
                    occupancy2 += (1 + (element-2))
            # Set up a particle selection by creating the Selection property:
            selection = data.particles_.create_property('Selection')

            # total number of types
            type_list = [x for x in range(1, len(self.mass_data)+1)]
            # This logic should work generall for any number of types
            # Lower triangular matrix of pairs excluding diagonal contains all unique pairs
            pair_matrix = np.array([[[i, j] for j in type_list] for i in type_list])
            i_lower_ex, j_lower_ex = np.tril_indices(len(type_list), k=-1)
            lower_triangular_matrix = pair_matrix[i_lower_ex, j_lower_ex] # Unique pairs

            for n in range(len(lower_triangular_matrix)):
                # Select A-sites occupied by exactly one B, C, or D atom
                # (the occupancy of the corresponding atom type must be 1, and all others 0)
                selection[...] |= ((site_type == lower_triangular_matrix[n][0]) & (occupancies[:, lower_triangular_matrix[n][1] - 1] == 1) & (total_occupancy == 1)) 
                             
                
            antisite_indices = np.where(selection == 1)[0]


            # Count the total number of antisite defects
            antisite_count = np.count_nonzero(selection[...])

            # Output the total number of antisites as a global attribute:
            data.attributes['Antisite_count'] = antisite_count
            tot_num_defects =  antisite_count + occupancy0 + occupancy2

            defect_count = tot_num_defects
        return defect_count

    def modify_lammps_in (self, filename, pka_vector):
        """
        Modifies the LAMMPS input file by setting the velocity of the PKA atom
        and the group ID of the PKA atom.

        Parameters:
            filename (str): The name of the LAMMPS input file to modify
            pka_vector (list of floats): The velocity vector of the PKA atom

        Returns:
            None
        """
        with open (filename, 'r') as f:
            lines = f.readlines()

        with open (filename, 'w') as f:
            for line in lines:
                if 'velocity PKA set' in line:
                    f.write(f'velocity PKA set {pka_vector[0]} {pka_vector[1]} {pka_vector[2]}\n')
                elif 'group PKA id' in line:
                    f.write(f'group PKA id {self.atom_id}\n')
                else:
                    f.write(line)


    def run_tde_loop (self, start_energy=None):
        """
        Runs a loop to search for the threshold defect energy (TDE).

        The loop starts at the given start_energy and increments the energy by 2 eV until a defect is found.
        Once a defect is found, the loop steps back down by 1 eV until no defect is found.
        The lowest energy with a defect found is returned as the TDE.

        Parameters:
            start_energy (float): The energy to start the search at (default: 5 eV)

        Returns:
            float: The lowest energy with a defect found (TDE)
        """
        start_energy = start_energy if start_energy is not None else 5
        energy_increment = 2
        max_energy = 1000

        energy = start_energy
        tde_energy = None
        while True:
            tempdirname = str(np.random.randint(1, 1000))
            try:
                os.mkdir(tempdirname)
                os.chdir(tempdirname)
                break 
            except FileExistsError:
                # Directory already exists, try again
                continue
        os.system('cp ../tde.in .')
        while energy < max_energy:
            
            # Calculate velocity and run simulation
            vel_vector = self.calculate_pka_vel(self.direction, pka_energy=energy)
            self.modify_lammps_in(filename='tde.in', pka_vector=vel_vector)
            subprocess.run(self.run_line, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            defect_count = self.defect_check()

            if defect_count != 0:  # defect found
                # Now step back down by 1 eV until no defect
                while defect_count != 0 and energy > 0:
                    energy -= 1
                    vel_vector = self.calculate_pka_vel(self.direction, pka_energy=energy)
                    self.modify_lammps_in(filename='tde.in', pka_vector=vel_vector)
                    subprocess.run(self.run_line, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    defect_count = self.defect_check()

                # Once we exit, energy is the lowest energy with a dfect found
                tde_energy = energy + 1
                break  # stop searching, we’ve bracketed the threshold

            else:
                energy += energy_increment  # keep increasing
                
        os.chdir('../')
        os.system(f'rm -rf {tempdirname}')

        return float(tde_energy)
                
    
def evaluate_tde(cfg, u: np.ndarray, start_energy = 5) -> Tuple[float, float]:

    """
    Evaluates the threshold defect energy (TDE) for a given direction `u`
    and LAMMPS configuration `cfg`.

    Parameters:
        cfg (Config): LAMMPS configuration
        u (np.ndarray): Direction of the PKA atom
        start_energy (float): Energy to start the search at (default: 5 eV)

    Returns:
        Tuple[float, float]: The TDE energy and error (0.5) for GP noise model.
    """
    energy = TDE_simulation(direction = u, run_line=cfg.run_line).run_tde_loop(start_energy = start_energy)
    return float(energy), 0.5









from setuptools import setup
setup(name='DEmap',
    version='0.5.0',
    url = 'https://github.com/Ash-Dickson/DEmap',
    description='Machine-learn the threshold displacement energy surface of a material with Gaussian Process Regression. ',
    author='Ashley Dickson',
    author_email='a.dickson2@lancaster.ac.uk',
    license='MIT license',
    packages=['DEmap'],
    install_requires=['gpytorch',
                      'numpy', 'torch', 'ovito', 'dataclasses', 'matplotlib', 'scipy']
)



import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # Needed for 3D projection
from sphere_utils import *
# Example: generate points in a wedge
pts = fibonacci_sphere(
    n=200,
    theta_range=(0, np.pi),   # colatitude 0..45 degrees
    phi_range=(0, np.pi*2)      # azimuth 0..90 degrees
)

# 3D plot
fig = plt.figure(figsize=(6,6))
ax = fig.add_subplot(111, projection='3d')

ax.scatter(pts[:,0], pts[:,1], pts[:,2], color='blue', s=20)

ax.set_xlabel('X')
ax.set_ylabel('Y')
ax.set_zlabel('Z')
ax.set_box_aspect([1,1,1])  # Equal aspect ratio

plt.show()

